@*****************************************
@* RNC2 Unpacking Lib for ARM/Thumb GCC  *
@*  by Jeff Frohwein                     *
@* http://www.devrs.com                  *
@*****************************************
@ v1.0228 - Original release
@ v1.0612 - Changed stmed/ldmed to the proper stmfd/ldmfd opcodes.
@           Fixed WRITE16 bug after _bytedisp label.
@ v1.0616 - Fixed WRITE16 bug after _bytelp label.

@  Destination RAM area must be an even number of bytes
@ if WRITE16 flag is set. Add 1 extra byte than needed
@ if necessary.

@  Destination pointer must be 2 byte aligned if
@ dest ptr >= 0x6000000 and if WRITE16 flag is set. An
@ abort will occur at the start if not properly aligned.

@ Unpack RNC2 data using ARM Asm
@ (You must use -m 2 method in the compressor.)
@ Entry:  srcptr = Source packed data pointer
@        destptr = Destination RAM pointer for unpacked data
@ (void) UnpackRNC2 (u8 *srcptr, u8 *destptr);

@ Uncomment the next line to support unpacking to 16-bit GBA VRAM.
@ It slows unpacking down to do that but that's the only way.
 .equ _WRITE16,1

@ EXAMPLE USAGE:
@  test.c -
@     typedef  unsigned char  u8;
@     #define _VRAM (u8 *)0x6000000
@     extern void UnpackRNC2(u8 *srcadr, u8 *dstadr);
@     extern u8 rncdata[];
@
@     UnpackRNC2(&rncdata[0], _VRAM);
@
@  data.s -
@         .GLOBL  rncdata
@     rncdata:
@         .byte 82,78,67,2,0,0,16,0,0,0,11,104,57,236,160,19
@         .....
@         .byte 5,1,30,0,7,0,5,56,56,108,108,196,3,0,236,236
@
@  ( To convert raw binary data to a format that can be used
@    in a .h or .s file, get B2X for DOS/linux from
@    http://www.devrs.com/gba/  [Apps/Misc section] )

@ Register Usage:
@  r0 = src packed data
@  r1 = dest unpacked data
@  r2 = store data
@  r3 = Carry Flag
@  r4 = 0x80000000
@  r5 = A
@  r6 = B
@  r7 = C
@  r8 = temp
@  r9 = temp
@  r10 = temp used by STORE

        .ARM
        .ALIGN
        .GLOBL  UnpackRNC2

@ store byte in r2 @ r1
 .ifdef _WRITE16
  .MACRO STORE
         cmps    r10,#0           @ Is this the first byte?
         addeq   r10,r2,r4        @ Yes, store it
         beq     .+16             @ Yes, now exit
         add     r10,r10,r2,lsl #8
         strh    r10,[r1],#2
         mov     r10,#0
         .ENDM
 .else
  .MACRO STORE
         strb    r2,[r1],#1
        .ENDM
 .endif

@ Bit8 of A -> Bit0 of r3
 .MACRO PROCESS_CARRY
         mov     r3,r5,lsr #8
         ands    r3,r3,#1
        .ENDM

@@@ START OF CODE @@@

UnpackRNC2:
        stmfd   r13!,{r0-r10,r14}

 .ifdef _WRITE16
        mov     r10,#0          @ Init no data to store
        mov     r4,#0x80000000

   @ If writing to VRAM, exit if not 2 byte aligned
        cmps    r1,#0x6000000   @ is dest address to work ram?
        bcc    _notest          @ yes, skip align test

        tsts    r1,#1           @ Is dest pointer aligned ?
        bne     _exit           @ no, so exit
_notest:
 .endif

        add     r0,r0,#18       @ Skip over unused 18 byte RNC2 header

        cmps    r0,r0           @ Set carry flag

        ldrb    r5,[r0],#1
        adcs    r5,r5,r5
        adds    r5,r5,r5
        b       _xloop

_fetch3:
        ldrb    r5,[r0],#1
        add     r5,r5,r5
        adds    r5,r5,r3
        PROCESS_CARRY
        b       _back3
_fetch4:
        ldrb    r5,[r0],#1
        add     r5,r5,r5
        adds    r5,r5,r3
        b       _back4
_fetch5:
        ldrb    r5,[r0],#1
        add     r5,r5,r5
        adds    r5,r5,r3
        PROCESS_CARRY
        b       _back5
_fetch6:
        ldrb    r5,[r0],#1
        add     r5,r5,r5
        adds    r5,r5,r3
        b       _back6
_fetch7:
        ldrb    r5,[r0],#1
        add     r5,r5,r5
        adds    r5,r5,r3
        PROCESS_CARRY
        b       _back7
_raw:
        mov     r7,#4
_x4bits:
        add     r5,r5,r5
        PROCESS_CARRY
        ands    r5,r5,#0xff
        beq     _fetch7
_back7:
        add     r6,r6,r6
        add     r6,r6,r3
        subs    r7,r7,#1
        bne     _x4bits

        add     r7,r6,#3
        add     r7,r7,r7        @ C = (B+3)*2
_rawlpb:
        ldrb    r2,[r0],#1
        STORE
        ldrb    r2,[r0],#1
        STORE

        subs    r7,r7,#1
        bne     _rawlpb

        b       _xloop

@ --------------------------------------------------------------

_fetch0:
        ldrb    r5,[r0],#1
        add     r5,r5,r5
        add     r5,r5,r3
        tsts    r5,#0x100
        bne     _smalls
_getlen:
        add     r5,r5,r5
        PROCESS_CARRY
        ands    r5,r5,#0xff
        beq     _fetch3
_back3:
        add     r7,r7,r7
        add     r7,r7,r3

        add     r5,r5,r5
        PROCESS_CARRY
        tsts    r5,#0xff
        beq     _fetch4
_back4:
        tsts    r5,#0x100
        beq     _copy

        add     r5,r5,r5
        PROCESS_CARRY
        ands    r5,r5,#0xff
        beq     _fetch5
_back5:
        sub     r7,r7,#1

        add     r7,r7,r7
        add     r7,r7,r3
        cmps    r7,#9

        beq     _raw

@ --------------------------------------------------------------

_copy:
        add     r5,r5,r5
        PROCESS_CARRY
        tsts    r5,#0xff
        beq     _fetch6
_back6:
        tsts    r5,#0x100
        beq     _bytedisp

        add     r5,r5,r5
        PROCESS_CARRY
        ands    r5,r5,#0xff
        bne     _skip0

        ldrb    r5,[r0],#1
        add     r5,r5,r5
        add     r5,r5,r3
        PROCESS_CARRY
_skip0:
        add     r6,r6,r6
        add     r6,r6,r3
        add     r5,r5,r5
        PROCESS_CARRY
        tsts    r5,#0xff
        bne     _skip1

        ldrb    r5,[r0],#1
        add     r5,r5,r5
        add     r5,r5,r3
_skip1:
        tsts    r5,#0x100
        bne     _bigdisp

        cmps    r6,#0
        bne     _bytedisp

        add     r6,r6,#1
_another:
        add     r5,r5,r5
        PROCESS_CARRY
        ands    r5,r5,#0xff
        bne     _dispx

        ldrb    r5,[r0],#1
        add     r5,r5,r5
        add     r5,r5,r3
        PROCESS_CARRY
_dispx:
        add     r6,r6,r6
        add     r6,r6,r3
_bytedisp:
 .ifdef _WRITE16
        cmps    r10,#0          @ Do we have an output byte in cache?
        strneh  r10,[r1]
        addne   r1,r1,#1        @ Yes, effectively update the out pointer
 .endif

        ldrb    r3,[r0],#1
        and     r8,r1,#0xff
        subs    r3,r8,r3         @ l = r1(Low) - [r0]
        and     r3,r3,#0xff
        mov     r8,r1,lsr #8
        sbc     r9,r8,r6         @ h = r1(high) - B
        add     r3,r3,r9,lsl #8
        sub     r3,r3,#1

 .ifdef _WRITE16
        cmps    r10,#0          @ Do we have an output byte in cache?
        subne   r1,r1,#1        @ Yes, effectively update the out pointer
 .endif

_bytelp:
 .ifdef _WRITE16
        cmps    r10,#0          @ Do we have an output byte in cache?
        strneh  r10,[r1]
 .endif
        ldrb    r2,[r3],#1
        STORE
        subs    r7,r7,#1
        bne     _bytelp

        b       _xloop

@ --------------------------------------------------------

_getbits:
        ldrb    r5,[r0],#1
        add     r5,r5,r5
        add     r5,r5,r3
        tsts    r5,#0x100
        bne     _string
_xbyte:
        ldrb    r2,[r0],#1
        STORE
_xloop:
        add     r5,r5,r5
        PROCESS_CARRY
        bne     _chkz

        ldrb    r2,[r0],#1
        STORE

        add     r5,r5,r5
        tsts    r5,#0x100
        beq     _xbyte
        mov     r3,#1
_chkz:
        ands    r5,r5,#0xff
        beq     _getbits

@ --------------------------------------------------------

_string:
        mov     r6,#0
        mov     r7,#2
        add     r5,r5,r5
        PROCESS_CARRY
        tsts    r5,#0xff
        beq     _fetch0

        tsts    r5,#0x100
        beq     _getlen
_smalls:
        add     r5,r5,r5
        PROCESS_CARRY
        tsts    r5,#0xff
        beq     _fetch1
_back1:
        tsts    r5,#0x100
        beq     _bytedisp

        add     r7,r7,#1
        add     r5,r5,r5
        PROCESS_CARRY
        tsts    r5,#0xff
        beq     _fetch2
_back2:
        tsts    r5,#0x100
        beq     _copy

        ldrb    r7,[r0],#1
        cmps    r7,#0
        beq     _overnout

        add     r7,r7,#8

        b       _copy

@ --------------------------------------------------------

_bigdisp:
        add     r5,r5,r5
        PROCESS_CARRY
        ands    r5,r5,#0xff
        bne     _skip2

        ldrb    r5,[r0],#1
        add     r5,r5,r5
        add     r5,r5,r3
        PROCESS_CARRY
_skip2:
        add     r6,r6,r6
        add     r6,r6,r3
        orr     r6,r6,#4
        add     r5,r5,r5
        PROCESS_CARRY
        tsts    r5,#0xff
        bne     _skip3

        ldrb    r5,[r0],#1
        add     r5,r5,r5
        add     r5,r5,r3
_skip3:
        tsts    r5,#0x100
        bne     _bytedisp
        b       _another

_fetch1:
        ldrb    r5,[r0],#1
        add     r5,r5,r5
        add     r5,r5,r3
        b       _back1
_fetch2:
        ldrb    r5,[r0],#1
        add     r5,r5,r5
        add     r5,r5,r3
        b       _back2
_overnout:
        add     r5,r5,r5
        PROCESS_CARRY
        tsts    r5,#0xff
        bne     _check4end

        ldrb    r5,[r0],#1
        add     r5,r5,r5
        add     r5,r5,r3
_check4end:
        tsts    r5,#0x100
        bne     _xloop

 .ifdef _WRITE16
@ Is there one byte to store in the cache?
        cmps    r10,#0           @ Is there 1 byte left?
@        movne   r2,#0            @ yes, write byte in cache
@        addne   r10,r10,r2,lsl #8
        strneh  r10,[r1]         @,#2
 .endif

_exit:
        ldmfd   r13!,{r0-r10,r14}
        bx      r14

@ store halfword in r10 @ r1

@ShowSt:
@        stmfd   r13!,{r0-r10,r14}
@
@        ldr     r0,_ShowAdr
@        str     r1,[r0]
@
@        ldr     r0,_ShowData
@        str     r10,[r0]
@
@        bl      ShowSt1
@        ldmfd   r13!,{r0-r10,r14}
@        bx      r14
@
@ShowSt1:
@        ldr     r0,_ShowStr
@        orr     r0,r0,#1
@        bx      r0
@
@
@_ShowStr:
@        .word   ShowStr
@_ShowAdr:
@        .word   ShowAdr
@_ShowData:
@        .word   ShowData

